% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bnns.R
\name{bnns_train}
\alias{bnns_train}
\title{Internal function for training the BNN}
\usage{
bnns_train(
  train_x,
  train_y,
  L = 1,
  nodes = rep(2, L),
  act_fn = rep(2, L),
  out_act_fn = 1,
  iter = 1000,
  warmup = 200,
  thin = 1,
  chains = 2,
  cores = 2,
  seed = 123,
  prior_weights = NULL,
  prior_bias = NULL,
  prior_sigma = NULL,
  verbose = FALSE,
  refresh = max(iter/10, 1),
  normalize = TRUE,
  ...
)
}
\arguments{
\item{train_x}{A numeric matrix representing the input features (predictors) for training. Rows correspond to observations, and columns correspond to features.}

\item{train_y}{A numeric vector representing the target values for training. Its length must match the number of rows in \code{train_x}.}

\item{L}{An integer specifying the number of hidden layers in the neural network. Default is 1.}

\item{nodes}{An integer or vector specifying the number of nodes in each hidden layer. If a single value is provided, it is applied to all layers. Default is 16.}

\item{act_fn}{An integer or vector specifying the activation function(s) for the hidden layers. Options are:
\itemize{
\item \code{1} for tanh
\item \code{2} for sigmoid (default)
\item \code{3} for softplus
\item \code{4} for ReLU
\item \code{5} for linear
}}

\item{out_act_fn}{An integer specifying the activation function for the output layer. Options are:
\itemize{
\item \code{1} for linear (default)
\item \code{2} for sigmoid
\item \code{3} for softmax
}}

\item{iter}{An integer specifying the total number of iterations for the Stan sampler. Default is \code{1e3}.}

\item{warmup}{An integer specifying the number of warmup iterations for the Stan sampler. Default is \code{2e2}.}

\item{thin}{An integer specifying the thinning interval for Stan samples. Default is 1.}

\item{chains}{An integer specifying the number of Markov chains. Default is 2.}

\item{cores}{An integer specifying the number of CPU cores to use for parallel sampling. Default is 2.}

\item{seed}{An integer specifying the random seed for reproducibility. Default is 123.}

\item{prior_weights}{A list specifying the prior distribution for the weights in the neural network.
The list must include two components:
\itemize{
\item \code{dist}: A character string specifying the distribution type. Supported values are
\code{"normal"}, \code{"uniform"}, and \code{"cauchy"}.
\item \code{params}: A named list specifying the parameters for the chosen distribution:
\itemize{
\item For \code{"normal"}: Provide \code{mean} (mean of the distribution) and \code{sd} (standard deviation).
\item For \code{"uniform"}: Provide \code{alpha} (lower bound) and \code{beta} (upper bound).
\item For \code{"cauchy"}: Provide \code{mu} (location parameter) and \code{sigma} (scale parameter).
}
}
If \code{prior_weights} is \code{NULL}, the default prior is a \code{normal(0, 1)} distribution.
For example:
\itemize{
\item \code{list(dist = "normal", params = list(mean = 0, sd = 1))}
\item \code{list(dist = "uniform", params = list(alpha = -1, beta = 1))}
\item \code{list(dist = "cauchy", params = list(mu = 0, sigma = 2.5))}
}}

\item{prior_bias}{A list specifying the prior distribution for the biases in the neural network.
The list must include two components:
\itemize{
\item \code{dist}: A character string specifying the distribution type. Supported values are
\code{"normal"}, \code{"uniform"}, and \code{"cauchy"}.
\item \code{params}: A named list specifying the parameters for the chosen distribution:
\itemize{
\item For \code{"normal"}: Provide \code{mean} (mean of the distribution) and \code{sd} (standard deviation).
\item For \code{"uniform"}: Provide \code{alpha} (lower bound) and \code{beta} (upper bound).
\item For \code{"cauchy"}: Provide \code{mu} (location parameter) and \code{sigma} (scale parameter).
}
}
If \code{prior_bias} is \code{NULL}, the default prior is a \code{normal(0, 1)} distribution.
For example:
\itemize{
\item \code{list(dist = "normal", params = list(mean = 0, sd = 1))}
\item \code{list(dist = "uniform", params = list(alpha = -1, beta = 1))}
\item \code{list(dist = "cauchy", params = list(mu = 0, sigma = 2.5))}
}}

\item{prior_sigma}{A list specifying the prior distribution for the \code{sigma} parameter in regression
models (\code{out_act_fn = 1}). This allows for setting priors on the standard deviation of the residuals.
The list must include two components:
\itemize{
\item \code{dist}: A character string specifying the distribution type. Supported values are
\code{"half-normal"} and \code{"inverse-gamma"}.
\item \code{params}: A named list specifying the parameters for the chosen distribution:
\itemize{
\item For \code{"half-normal"}: Provide \code{sd} (standard deviation of the half-normal distribution).
\item For \code{"inverse-gamma"}: Provide \code{shape} (shape parameter) and \code{scale} (scale parameter).
}
}
If \code{prior_sigma} is \code{NULL}, the default prior is a \code{half-normal(0, 1)} distribution.
For example:
\itemize{
\item \code{list(dist = "half_normal", params = list(mean = 0, sd = 1))}
\item \code{list(dist = "inv_gamma", params = list(alpha = 1, beta = 1))}
}}

\item{verbose}{TRUE or FALSE: flag indicating whether to print intermediate output from Stan on the console, which might be helpful for model debugging.}

\item{refresh}{refresh (integer) can be used to control how often the progress of the sampling is reported (i.e. show the progress every refresh iterations). By default, refresh = max(iter/10, 1). The progress indicator is turned off if refresh <= 0.}

\item{normalize}{Logical. If \code{TRUE} (default), the input predictors
are normalized to have zero mean and unit variance before training.
Normalization ensures stable and efficient Bayesian sampling by standardizing
the input scale, which is particularly beneficial for neural network training.
If \code{FALSE}, no normalization is applied, and it is assumed that the input data
is already pre-processed appropriately.}

\item{...}{Currently not in use.}
}
\value{
An object of class \code{"bnns"} containing the following components:
\describe{
\item{\code{fit}}{The fitted Stan model object.}
\item{\code{call}}{The matched call.}
\item{\code{data}}{A list containing the Stan data used in the model.}
}
}
\description{
This function performs the actual fitting of the Bayesian Neural Network.
It is called by the exported bnns methods and is not intended for direct use.
}
\details{
The function uses the \code{generate_stan_code} function to dynamically generate Stan code based on the specified number of layers and nodes. Stan is then used to fit the Bayesian Neural Network.
}
\examples{
\donttest{
# Example usage:
train_x <- matrix(runif(20), nrow = 10, ncol = 2)
train_y <- rnorm(10)
model <- bnns::bnns_train(train_x, train_y,
  L = 1, nodes = 2, act_fn = 2,
  iter = 1e1, warmup = 5, chains = 1
)

# Access Stan model fit
model$fit
}
}
\seealso{
\code{\link[rstan]{stan}}
}
\keyword{internal}
